#!/usr/bin/env python3
# coding: utf-8

_closing_of_known_tags = {
    'a': 2,
    'img': 0,
    'script': 2,
    'link': 0,
    'source': 0,
    'audio': 0,
    'video': 0,
}

_url_attrname_of_known_tags = {
    'a': 'href',
    'link': 'href',
    'img': 'src',
    'script': 'src',
    'source': 'src',
    'audio': 'src',
    'video': 'src',
}

_default_props_of_known_tags = {
    'img': {
        'alt': '',
    },
    'script': {
        'type': "text/javascript",
    },
    'link': {
        'rel': "stylesheet",
        'type': "text/css",
    }
}

_templates = {
    0: '<{tagname} {props}>',
    1: '<{tagname} {props} />',
    2: '<{tagname} {props}>{content}</{tagname}>',
}


def fmttag(name, content, props=None, closing=None):
    props_parts = []
    for k, v in (props or {}).items():
        p = '{}="{}"'.format(k, v)
        props_parts.append(p)
    props_string = ' '.join(props_parts)
    ctx = {'tagname': name, 'props': props_string}
    if closing is None:
        closing = _closing_of_known_tags.get(name, 2)
    if closing == 2:
        ctx['content'] = content
    return _templates[closing].format(**ctx)


def wrap(content, *names):
    """
    :param content: string
    :param names: e.g. ('html', 'body')
    :return: string
    """
    parts = []
    names = list(names)
    for x in names:
        parts.append('<{}>'.format(x))

    parts.append(content)
    names.reverse()
    for x in names:
        parts.append('</{}>'.format(x))
    return ''.join(parts)


def wrap_url(url, tagname):
    props = _default_props_of_known_tags.get(tagname, {})
    props[_url_attrname_of_known_tags.get(tagname, 'href')] = url
    return fmttag(tagname, '', props=props)


def run_wrap_url(_, args):
    import sys
    tagname = args[0] if args else 'a'
    for line in sys.stdin:
        print(wrap_url(line.strip(), tagname))


def wrap_default(content):
    return wrap(content, 'html', 'body')


def table(rows, props=None):
    trs = []
    for row in rows:
        tr = ''.join("<td>" + c + "</td>" for c in row)
        tr = '<tr>' + tr + '</tr>'
        trs.append(tr)
    return fmttag('table', '\n'.join(trs), props)



#!/usr/bin/env python3
# coding: utf-8

import base64
import logging
import os

from joker.environ.utils import *  # noqa


# deprecated. use joker.environ.utils.check_{inclusive,exclusive}_* functions
def check_prefixes(string: str, include=None, exclude=None) -> bool:
    """
    Args:
        string (str): string to be checked
        include: None = include anything
        exclude: None = no exlusion rule
    """
    if exclude is not None:
        for prefix in exclude:
            if string.startswith(prefix):
                return False
    if include is not None:
        for prefix in include:
            if string.startswith(prefix):
                return True
        return False
    else:
        return True


def easylog(level):
    root = logging.root
    fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(fmt)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    root.setLevel(level)
    root.addHandler(handler)


def read_lines(path: str):
    for line in open(path):
        line = line.strip()
        if not line:
            continue
        yield line


# Python 3.5+
if hasattr(bytes, 'hex'):
    def random_hex(length=12):
        return os.urandom(length).hex()
else:
    def random_hex(size=12):
        b = os.urandom(size)
        return base64.b16encode(b).decode('ascii')



#!/usr/bin/env python3
# coding: utf-8

from __future__ import unicode_literals

import hashlib
import re
import zlib

import six

from joker.stream.utils import checksum


def guess_hash_algorithm(digest):
    algo_map = {
        ('bin', 16): 'md5',
        ('bin', 20): 'sha1',
        ('bin', 28): 'sha224',
        ('bin', 32): 'sha256',
        ('bin', 48): 'sha384',
        ('bin', 64): 'sha512',
        ('hex', 32): 'md5',
        ('hex', 40): 'sha1',
        ('hex', 56): 'sha224',
        ('hex', 64): 'sha256',
        ('hex', 96): 'sha384',
        ('hex', 128): 'sha512',
    }
    if isinstance(digest, six.binary_type):
        try:
            digest = digest.decode('utf-8')
        except Exception:
            return algo_map.get(('bin', len(digest)))
    if not re.match(r'[0-9A-Fa-f]+', digest):
        return
    return algo_map.get(('hex', len(digest)))


class HashedPath(object):
    """
    /data/web/config.ini;;f6da8154d73f954cdfebd04d6c76cff8
    for integrity-check purpose only
    """
    delimiter = ';;'

    def __init__(self, digest, algo, path):
        self.digest = digest
        self.algo = algo
        self.path = path

    @staticmethod
    def compute_hash(path, algo='md5'):
        try:
            return checksum(path, algo=algo).hexdigest()
        except IOError:
            return hashlib.new(algo, b'\0').hexdigest()

    @classmethod
    def generate(cls, path, algo='md5'):
        digest = cls.compute_hash(path, algo)
        return cls(digest, algo, path)

    @classmethod
    def parse(cls, h_path):
        if cls.delimiter not in h_path:
            return cls('', 'md5', h_path)
        path, digest = h_path.split(cls.delimiter, 2)
        algo = guess_hash_algorithm(digest)
        return cls(digest, algo, path)

    def __str__(self):
        return '{}{}{}'.format(self.path, self.delimiter, self.digest)

    def verify(self):
        if not self.digest:
            return True
        hx1 = self.generate(self.path, self.algo or 'md5')
        return hx1.digest == self.digest


class Hasher(object):
    """
    Unified interface for hashlib & zlib hash functions
    """
    __hfuncs__ = {
        'adler32': zlib.adler32,
        'crc32': zlib.crc32,
        'md5': hashlib.md5,
        'sha1': hashlib.sha1,
        'sha224': hashlib.sha224,
        'sha256': hashlib.sha256,
        'sha384': hashlib.sha384,
        'sha512': hashlib.sha512,
    }

    __hlens__ = {
        32: 'md5',
        40: 'sha1',
        56: 'sha224',
        64: 'sha256',
        96: 'sha384',
        128: 'sha512',
    }

    def __init__(self, name):
        self.name = name.lower()
        if self.name not in self.__hfuncs__:
            raise ValueError("hash func '{}' not found".format(name))
        if name in {'adl32', 'crc32'}:
            self.hash = getattr(zlib, name)  # a checksum function
            self.emulate = True
            self.result = self.hash(bytes())  # intermediate result
        else:
            self.hash = getattr(hashlib, name)()  # a hash object
            self.emulate = False
            self.result = 0  # not used for hashlib functions

    def __repr__(self):
        cn = self.__class__.__name__
        return '{}({})'.format(cn, repr(self.name))

    def _update_with_bytes(self, data):
        """
        :param data: (bytes)
        """
        # hashlib.md* / hashlib.sha*
        if not self.emulate:
            self.hash.update(data)
            return
        # zlib.adler32 / zlib.crc32
        if self.result is None:
            self.result = self.hash(data)
        else:
            self.result = self.hash(data, self.result)

    def _update_with_traverser(self, data):
        # from joker.nested.traverse import Traverser
        # traverser = Traverser(self, lambda x: self.update(x))
        # traverser()
        raise NotImplementedError

    def update(self, data):
        if isinstance(data, six.text_type):
            return self._update_with_bytes(data.encode())
        elif isinstance(data, six.binary_type):
            return self._update_with_bytes(data)
        self._update_with_traverser(data)

    def hexdigest(self):
        # hashlib funcs
        if not self.emulate:
            return self.hash.hexdigest()
        # zlib.adler32 / zlib.crc32
        return hex(self.result & 0xffffffff).replace('0x', '')



#!/usr/bin/env python3
# coding: utf-8

import os

import joker.environ


class GlobalInterface(joker.environ.GlobalInterface):
    package_name = 'm14.environ'

    def under_data_dir(self, *paths, mkdirs=False):
        if 'data_dir' not in self.conf:
            names = self.package_name.split('.')
            data_dir = os.path.join('/data/local', *names)
            self.conf.setdefault('data_dir', data_dir)
        return super().under_data_dir(*paths, mkdirs=mkdirs)


#!/usr/bin/env python3
# coding: utf-8

import json
import re
import shlex
import sys
from functools import lru_cache

from joker.textmanip import random_hex
# TODO: drop this
from joker.textmanip.path import make_new_path
from volkanic.introspect import razor

_symbols = [random_hex]


def dump_json_request_to_curl(method: str, url: str, data=None, aslist=False):
    method = method.upper()
    if method == 'GET':
        parts = ['curl', url]
    else:
        parts = [
            'curl', '-X', method, url,
            '-H', 'Content-Type: application/json',
            '-d', json.dumps(razor(data), ensure_ascii=False),
        ]
    if aslist:
        return parts
    parts = [shlex.quote(s) for s in parts]
    return ' '.join(parts)


@lru_cache(100)
def get_json_config(path):
    return json.load(open(path))


def copy_fields(record: dict, keys: list, keymap: dict, default=None):
    new_record = {k: record.get(k, default) for k in keys}
    for old_key, new_key in keymap.items():
        new_record[new_key] = record.get(old_key, default)
    return new_record


def camel_case_split(name: str):
    # https://stackoverflow.com/a/29920015/2925169
    matches = re.finditer(
        '.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)', name
    )
    return [m.group(0) for m in matches]


def camel_case_to_underscore(name: str):
    return '_'.join(s.lower() for s in camel_case_split(name))


def text_routine_catstyle(func, *paths):
    # routine function for quick and dirty text manipulation scripts
    if not paths:
        paths = sys.argv[1:]

    if not paths:
        for line in sys.stdin:
            print(func(line))
        return
    for path in paths:
        for line in open(path):
            print(func(line))


def text_routine_perfile(func, ext='.out'):
    # routine function for quick and dirty text manipulation scripts
    for path in sys.argv[1:]:
        outpath = make_new_path(path, ext)
        func(path, outpath)


def chained_bracket(record: dict, *keys):
    current_rec = record
    for k in keys:
        try:
            current_rec = current_rec[k]
        except (KeyError, IndexError):
            current_rec = {}
    return current_rec


def dotget(record: dict, key: str):
    return chained_bracket(record, key.split('.'))


def _get_instance_from_method_args(func, args: tuple):
    if not args:
        return
    instance = args[0]
    try:
        raw_func = instance.__class__.__dict__.get(func.__name__).raw_func
    except (AttributeError, TypeError):
        return
    if raw_func != func:
        return
    return instance


def once_per_instance(func):
    """Runs a method (successfully) only once per instance."""

    @wraps(func)
    def _resulting_method(*args, **kwargs):
        self = _get_instance_from_method_args(func, args)
        d = getattr(self, '__dict__', {})
        if d.get(func.__qualname__):
            msg = f'method {func.__qualname__} can only be called once'
            raise RuntimeError(msg)
        rv = func(*args, **kwargs)
        d[func.__qualname__] = 1
        return rv

    _resulting_method.raw_func = func
    return _resulting_method
